"""
Author: Thomas M. Lloyd
Master's thesis Spring 2021
"""

import numpy as np
from gurobipy import *
import pandas as pd
from pandas import ExcelWriter
from openpyxl import load_workbook
import xlrd as xl
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

def create_gant(result_list, vessels, n_turbines, n_operations=7, end_time=110):

    # Declaring a figure "gnt"
    fig, gnt = plt.subplots()

    # Setting Y-axis limits
    gnt.set_ylim(0, 40)

    # Setting X-axis limits
    gnt.set_xlim(0, end_time+30)

    # Setting labels for x-axis and y-axis
    gnt.set_xlabel('Number of days')
    gnt.set_ylabel('Operations')

    # Setting ticks on y-axis
    gnt.set_yticks([5, 10, 15, 20, 25, 30, 35])
    # Labelling ticks of y-axis
    gnt.set_yticklabels(['Final Commissioning', 'Cable Installation','Hook-up',
                         'Tow Out','Sea-keeping','Anchor Installation','Anchor Transportation'])

    # Setting graph attribute
    gnt.grid(True)
    list_color = cm.get_cmap('ocean', 15)

    # Declaring bars
    for i in range(len(result_list)):
        operation_id = result_list[i][0]
        start = result_list[i][2]
        duration = result_list[i][3] - result_list[i][2]  
        vessel = result_list[i][4]
        color = vessel / (len(vessels)+4)  # Find a normalized value for color definition
        bar_height = 4
        bar_space = 5
        y_size = 38

        if (operation_id * bar_space) % 35 == 0:
            increment = 35
        else:
            increment = (bar_space * operation_id) % 35

        gnt.broken_barh([(start, duration)],
                        ((y_size - increment),
                        bar_height),
                        facecolors=list_color(color),
                        label=vessel)

    gnt.legend()
    handles, labels = gnt.get_legend_handles_labels()
    lgd = dict(zip(labels, handles))
    gnt.legend(lgd.values(), lgd.keys())
    plt.savefig('TenTurbinesCFFSHs2.png',bbox_inches='tight')
    plt.show()
    
    return gnt


def create_solution_list(m, vessel_job, job_time, nodes, vessels, operation_time, sail_time, wait_time, cost_TC):

    List = []
    # Print solution
    if m.status == GRB.Status.OPTIMAL:
        solution = m.getAttr('x', vessel_job)
        solution2 = m.getAttr('x', wait_time)
        solution3 = m.getAttr('x', job_time)

        # print(solution)
        start_times = []
        end_times = []
        vessel_selected = []
        wait_times=[]
        total_cost=[]
        for i in nodes:
            for v in vessels:
                for j in nodes:
                    if solution[i, j, v] > 0:
                        #print(solution[i, j, v])
                        operation_end = solution3[i, v] + operation_time[v][i] + sail_time[v][i]
                        tot_cost = ((operation_time[v][i] + sail_time[v][i])*solution[i, j, v]*cost_TC[v])
                        value_list = []
                        if i != 0: #and i != nodes[-1]:
                            start_times.append(solution3[i, v])
                            wait_times.append(solution2[v])
                            end_times.append(operation_end)
                            vessel_selected.append(v)
                            total_cost.append(tot_cost)

                            value_list.append(i)
                            value_list.append(solution2[v])
                            value_list.append(solution3[i, v])
                            value_list.append(operation_end)
                            value_list.append(v)
                            List.append(value_list)
      
    return List, start_times, end_times, vessel_selected, wait_times, total_cost


