"""
Author: Thomas Magnus Lloyd
Master's thesis Spring 2021
"""

import numpy as np
from gurobipy import *
import pandas as pd
from pandas import ExcelWriter
from openpyxl import load_workbook
import xlrd as xl
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
from Functions import *
from itertools import repeat

# Dataframes
dfV = pd.read_excel("Vessel_Info.xlsx", sheet_name='Vessels')
dfVC = pd.read_excel("Vessel_Info.xlsx", sheet_name='Act_req', index_col=0)
dfVN = pd.read_excel("Vessel_Info.xlsx", sheet_name='Dur_Vessel', index_col=0)
dfST = pd.read_excel("Vessel_Info.xlsx", sheet_name='Sail_vessel', index_col=0)
data = pd.read_excel("Vessel_Info.xlsx", sheet_name='weather')

solutions = []
# Sets
for n_turbines in[2]:
    n_duplicates = 1
    n_operations = 7
    size_input = dfV.shape[0]
    
    
    vessel_ids = np.arange(size_input)
    vessels = np.arange(n_duplicates * size_input)
    operations = np.arange(1, n_turbines * n_operations + 1)
    nodes = np.arange(n_turbines * n_operations + 2)
    
    # Vertices
    vertices_to = []
    to = []
    vertices_been = []
    been = []
    for r in nodes:
        to = np.array(range(r + 1, len(nodes)))
        been = np.array(range(0, r))
        vertices_to.append(to)
        vertices_been.append(been)
        to = []
        been = []
    
    # Creating arc set
    Av = [(1, 2), (2, 5), (3, 4), (4, 5), (5, 6), (6, 7), (1, 8)]
    for i in range(1, n_turbines):
        for j in range(n_operations):
            tuple = (Av[j][0] + (n_operations * i), Av[j][1] + (n_operations * i))
            Av.append(tuple)
    Av = Av[:-1]
    
    # ----------------- PARAMETERS -------------------------
    # Start time and end time
    time_increment_start = 10
    time_increment_end = 10
    slack = 40
    start_time = [0, 0, 1, 2, 3, 4, 5, 6]
    end_time = [time_increment_start * n_turbines + slack, 6,8,10,12,16,18,20]
    
    # Weather calculations
    swh = ['sw2']
    df = pd.read_excel("Vessel_Info1.xlsx", sheet_name='Markov')

    df = df.dropna()
    df = df.sw2.tolist()
    k = 7
    NrHs = sum(i > k for i in df)
    time_increment_wait = 0
    count=[]
    count.extend(repeat(NrHs,8))
    
    for i in df[:50]:
        if i >= k :
            start_time = [0, 5, 6, 7, 8, 9, 10, 11]
            count = count
            end_time = [time_increment_start * n_turbines + slack, 14, 16, 18, 20, 24, 26, 28]
            time_increment_wait = 5
        else:
            start_time = start_time
            count = [0, 0, 0, 0, 0, 0, 0, 0]
            end_time = end_time
            time_increment_wait = time_increment_wait
    for i in range(1, n_turbines):
        for j in range(n_operations):
            start_time.append(start_time[j + 1] + time_increment_start * n_turbines)
            count.append(count[j+1])
            end_time.append(end_time[j + 1] + time_increment_end * n_turbines)
    start_time.append(time_increment_start * n_turbines)
    count.append(time_increment_wait * n_turbines)
    end_time.append(time_increment_end * n_turbines + slack)
    
    # Cost for chartering, mobilization and waiting
    cost_TC = dfV.Day_rate.tolist() * n_duplicates
    cost_mobilization = dfV.Mob_cost.tolist() * n_duplicates
    cost_W = dfV.Wait_cost.tolist() * n_duplicates
    bigM = 100000
    
        # Operation requirements
    requirement = np.asarray(dfVC)
    operation_requirement = []
    for j in vessel_ids:
        req = []
        if n_duplicates > 1:
            for n in range(n_duplicates):
                req = []
                for i in range(n_turbines):
                    req.extend(requirement[j][1:-1])
                operation_requirement.append(req)
        else:
            for i in range(0, n_turbines):
                req.extend(requirement[j][1:-1])
            operation_requirement.append(req)

    # Execution times
    time = np.asarray(dfVN)
    operation_time = []
    for j in vessel_ids:
        t = []
        if n_duplicates > 1:
            for n in range(n_duplicates):
                t = []
                t.append(0)
                for i in range(n_turbines):
                    t.extend(time[j][1:-1])
                t.append(0)
                operation_time.append(t)
        else:
            t.append(0)
            for n in range(n_turbines):
                t.extend(time[j][1:-1])
            t.append(0)
            operation_time.append(t)
    
    # Sailing time
    time1 = np.asarray(dfST)
    sail_time = []
    for j in vessel_ids:
        t = []
        if n_duplicates > 1:
            for n in range(n_duplicates):
                t = []
                t.append(0)
                for i in range(n_turbines):
                    t.extend(time1[j][1:-1])
                t.append(0)
                sail_time.append(t)
        else:
            t.append(0)
            for n in range(n_turbines):
                t.extend(time1[j][1:-1])
            t.append(0)
            sail_time.append(t)
            
    # ------------- MODEL CREATION -------------------
    # Create optimization model
    m = Model('schedule_optimization')
    
     # ------------- VARIABLES ------------------------
    vessel_job = m.addVars(nodes, nodes, vessels, vtype=GRB.BINARY, name="vessel_job")
    job_time = m.addVars(nodes, vessels, vtype=GRB.INTEGER, name="job_time")
    wait_time = m.addVars(vessels, vtype= GRB.INTEGER, name ="waiting_time")
    
    # --------------- OBJECTIVE FUNCTION ------------------
    # Operation time minimization (OT)
    m.setObjective(sum(cost_mobilization[v] * vessel_job[0, j, v] for v in vessels for j in nodes[:-1]) +
                   sum(vessel_job[i, j, v] * cost_TC[v] * (operation_time[v][i] + sail_time[v][i]) for v in vessels
                       for i in nodes for j in nodes)  + sum(cost_W[v]*wait_time[v] for v in vessels), GRB.MINIMIZE)
    
    # ------------- CONSTRAINTS --------------------------
    for i in operations:
        m.addConstr(sum(vessel_job[i, j, v] for v in vessels for j in vertices_to[i]) == 1, "one_task2")
        
    # Weather constraint
    for v in vessels:
        m.addConstr(
        wait_time[v] == (sum(vessel_job[i, j, v] * count[i] for i in nodes for j in nodes)), "wait_time")
         
    # Routing constraints
    for v in vessels:
        m.addConstr(sum(vessel_job[0, j, v] for j in vertices_to[0]) == 1, "operation_start")
    

    for v in vessels:
        m.addConstr(sum(vessel_job[i, nodes[-1], v] for i in vertices_been[nodes[-1]]) == 1, "operation_end")
        
        
    for j in operations:
        for v in vessels:
            m.addConstr(sum(vessel_job[i, j, v] for i in vertices_been[j]) -
                        sum(vessel_job[j, i, v] for i in vertices_to[j]) == 0, "routing constraint")
            
    # Time Constraint
    for i in nodes:
        for j in nodes:
            for v in vessels:
                m.addConstr(job_time[i, v] + operation_time[v][i] + sail_time[v][i] - job_time[j, v] -
                            bigM * (1 - vessel_job[i, j, v]) <= 0, "time_constraint")       
                        
    # Origin Time constraint
    for v in vessels:
        m.addConstr(start_time[0] <= job_time[0, v], "origin_time_constraint")
        m.addConstr(job_time[0, v] <= end_time[0], "origin_time_constraint1")

    for v in vessels:
        for i in operations:
            m.addConstr(start_time[i] * sum(vessel_job[i, j, v] for j in vertices_to[i]) <= job_time[i, v],
                        "time_windows")
            m.addConstr(job_time[i, v] <=
                        (end_time[i] - operation_time[v][i]) * sum(vessel_job[i, j, v] for j in vertices_to[i]),
                        "time_windows1")
            
    # Precedence constraints
    for i, j in Av:
        m.addConstr(sum(job_time[i, v] for v in vessels) +
                    sum((operation_time[v][i]+ sail_time[v][i]) * vessel_job[i, r, v] for r in vertices_to[i] for v in vessels)
                    - sum(job_time[j, v] for v in vessels) <= 0, "Precedence")
        
    # Equipment requirement constraint
    for v in vessels:
        for i in operations:
            m.addConstr(sum(vessel_job[i, j, v] for j in nodes) <= bigM * operation_requirement[v][i-1],
                        "Task_requirement")
    
    # Compute optimal solution
    m.optimize()
    obj = m.getObjective()
    best_objective = obj.getValue()
    runtime = m.Runtime
    MIPGap = m.MIPGap
    
# Create solution list
List, start_times, end_times, vessel_selected, wait_times, total_cost = create_solution_list(m, vessel_job, job_time, nodes, vessels, operation_time, sail_time, wait_time, cost_TC)


# Solutions of cost per turbine
solutions.append(best_objective / n_turbines)

# Creates a Gant chart
create_gant(List, vessels, n_turbines, n_operations=7, end_time=end_time[-1])